package de.fraunhofer.iem.dataprovider.taskManager

import de.fraunhofer.iem.dataprovider.logger.getLogger
import de.fraunhofer.iem.dataprovider.taskManager.tasks.GetGitlabProjectTask
import de.fraunhofer.iem.dataprovider.taskManager.tasks.ITask
import de.fraunhofer.iem.dataprovider.taskManager.tasks.TaskType
import jakarta.annotation.PreDestroy
import kotlinx.coroutines.*
import kotlinx.coroutines.channels.actor
import org.springframework.stereotype.Component

/**
 * The Task Manager takes tasks and distributes them to
 * underlying workers. Internally it uses a channel
 * to manage incoming tasks.
 */
@Component
class TaskManager: IWorker {

    // The used default dispatcher is ok for CPU-bound workloads. However,
    // if they block for a long time it's better to use a custom thread
    // pool solution.
    private val worker = Worker("Worker")
    // Should be used for long-running tasks, which DON'T use CPU power and
    // are non-blocking
    private val ioWorker = Worker("IO-Worker", CoroutineScope(Dispatchers.IO))
    // Should be used for long-running CPU heavy tasks, which potentially block.
    // private val customWorker = getCustomThreadpoolWorker()

    private val logger = getLogger(javaClass)
    private val mainScope = CoroutineScope(Dispatchers.Default)
    private val actor = mainScope.taskActor()
    override suspend fun addTask(task: ITask) {
            actor.send(task)
    }

    private fun CoroutineScope.taskActor() = actor<ITask> {
        for (task in channel) {
            // For test purposes we have the Done task, which prints a control
            // message printed in the task manager.
            if (task.type == TaskType.DONE) {
                task.execute()
            }
            logger.info("[${Thread.currentThread().name}] add task called in Task Manager $task")
            when(task) {
                is GetGitlabProjectTask -> ioWorker.addTask(task)
                else -> worker.addTask(task)
            }
            logger.info("[${Thread.currentThread().name}] add task finished in Task Manager $task")
        }
    }

    @PreDestroy
    override fun close() {
        worker.close()
    }
}
